/*
** Module   : AbstractPatternWorker.java
** Abstract : centralizes the most common init/term logic for pattern workers
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 GES 20050228   @20069 Created initial version supporting a
**                           resolver member, library registration and
**                           a default implementation of the
**                           PatternWorker interface.
** 002 ECF 20050228   @20090 Added methods to retrieve source and target
**                           ASTs currently being inspected.
** 003 ECF 20050310   @20272 Added default implementation of the
**                           resolveToLiteral method. Simply returns null.
** 004 ECF 20050315   @20345 Added default implementation of the visitAst
**                           method. Does nothing (required by
**                           PatternWorker interface).
** 005 ECF 20050328   @20515 Changed getTarget method to getCopy.
** 006 ECF 20050429   @20964 Changes related to namespace support for user
**                           functions and variables. Changed method
**                           signature for initialize and finish methods
**                           to match changes to PatternWorker interface.
**                           Pass this object as the namespace key when
**                           registering libraries.
** 007 ECF 20050525   @21290 Changes to support new expression engine
**                           implementation. Only a single library per
**                           pattern worker is now supported. Library
**                           registration now occurs in RuleContainer.
** 008 ECF 20080122   @37466 Plugged possible memory leak. Release resolver in finish().
** 009 ECF 20080730   @39263 Fixed possible memory leak. Release library in finish().
** 010 GES 20090518   @42383 Import change.
** 011 ECF 20130806          Added registerTree to register an arbitrary AST with the resolver.
*/

package com.goldencode.p2j.pattern;

import com.goldencode.ast.*;

/**
 * Provides a default implementation of the most common initialization and 
 * termination logic for pattern engine worker classes.  This class is
 * abstract since it cannot do anything useful on its own.  It only exists
 * as an optional parent class for other pattern workers, to allow them to
 * be written without duplicating the most common logic that most (if not
 * all) workers would have to implement.
 * <p>
 * The most important usage note is that it is not possible for sub-classes
 * to pass the <code>libraries</code> array of <code>Object</code> instances 
 * (the list of exported libraries to register) in the constructor of a
 * super class.  For this reason, <b>a separate <code>setLibraries</code>
 * method has been provided which must be called in the sub-class'
 * constructor, after the <code>super</code> constructor has been
 * invoked.</b>
 */
public abstract class AbstractPatternWorker
      implements PatternWorker
{
   /**
    * Provides access to source/target AST nodes and is our interface for
    * registering libraries and functions.
    */
   protected AstSymbolResolver resolver = null;

   /** Library object (exports functions/variables to rulesets) to register */
   private Object library = null;

   /**
    * Default constructor, which can only be called by a sub-class.
    */
   protected AbstractPatternWorker()
   {
   }

   /**
    * Obtains the reference to the <code>resolver</code> and initializes 
    * all exported libraries.
    */
   public void initialize()
   {
      resolver = AstSymbolResolver.getResolver();
   }

   /**
    * Register the given tree with the AST resolver. This will register the AST and all its
    * descendants.
    *
    * @param   ast
    *          Tree to register.
    */
   public void registerTree(Aast ast)
   {
      resolver.registerTree(ast);
   }

   /**
    * Hook to provide termination processing for a pattern worker.
    */
   public void finish()
   {
      resolver = null;
      library = null;
   }

   /**
    * Get the user function/property library exposed by this pattern worker.
    *
    * @see com.goldencode.p2j.pattern.PatternWorker#getLibrary()
    */
   public Object getLibrary()
   {
      return library;
   }

   /**
    * This method is called each time the pattern engine needs to resolve
    * a string constant into a numeric, boolean, or string literal value.
    * <p>
    * It must return an object of the following type (or <code>null</code>):
    * <ul>
    * <li>Long
    * <li>Double
    * <li>Boolean
    * <li>String
    * </ul>

    * @param   constant
    *          A string representing a constant to be resolved.
    *
    * @return  This default implementation always returns <code>null</code>.
    */
   public Object resolveConstant(String constant)
   {
      return null;
   }

   /**
    * This method is called each time a new AST is visited, just before the
    * pattern engine processes that AST. At the time this method is invoked,
    * the resolver is not yet configured with the any source or target AST
    * node, so the source AST is passed as a parameter, in case the worker
    * needs to extract information. In general, the worker should not modify
    * this AST. The default implementation does nothing.
    *
    * @param   ast
    *          The root node of the source AST about to be processed by the
    *          pattern engine.
    */
   public void visitAst(Aast ast)
   {
   }

   /**
    * Sets the library that <b>must be registered</b>.  This can only be
    * called by a subclass. This <b>must be called from within the 
    * subclass' constructor.</b>
    *
    * @param    library
    *           Instance of exported libraries to be registered with the
    *           <code>resolver</code>.  <b>Must not</b> be <code>null</code>.
    *
    * @throws   IllegalArgumentException
    *           if <code>library</code> is <code>null</code>.
    */
   protected void setLibrary(Object library)
   {
      if (library == null)
      {
         throw new IllegalArgumentException("No library to register");
      }

      this.library = library;
   }

   /**
    * Get the current source AST. If called while the pattern engine is
    * processing init- or post-rules, this will always be the root of the
    * source AST. If called while the pattern engine is walking the tree,
    * this will be the source node currently being inspected.
    *
    * @return  The AST node from the source tree currently under inspection.
    */
   protected Aast getSource()
   {
      return resolver.getSourceAst();
   }

   /**
    * Get the current copy AST. If called while the pattern engine is
    * processing init- or post-rules, this will always be the root of the
    * copy AST. If called while the pattern engine is walking the tree,
    * this will be the copy of the source node currently being inspected
    * Note that this could be a node which was removed from the tree by
    * some previous action.
    *
    * @return  The writable copy of the source AST node currently under
    *          inspection.
    */
   protected Aast getCopy()
   {
      return resolver.getCopyAst();
   }
}
