/*
** Module   : ConnectionManager.java
** Abstract : Provides services for logical database connections and aliases
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20051202   @23667 Created initial version. Provides services to
**                           manage logical database connections and
**                           aliases. Provides runtime support for the
**                           following Progress statements/functions:
**                              CONNECT/DISCONNECT
**                              CONNECTED
**                              CREATE/DELETE ALIAS
**                              ALIAS
**                              NUM-ALIASES
**                              LDB-/PDB-/SDB-/DBNAME
**                              NUM-DBS
**                              DBTYPE/DBRESTRICTIONS
** 002 ECF 20060120   @24038 Integration with P2J server. Replaced
**                           ThreadLocal with ContextLocal.
** 003 GES 20060517   @26198 Match new Finalizable interface requirement.
** 004 ECF 20060628   @27670 Fixed closeBufferScope. Only remove a buffer
**                           from the set of those which reference a
**                           physical database if its outermost scope is
**                           closing.
** 005 ECF 20070907   @35292 Fixed constructor to avoid adding the temp
**                           table dedicated database to the list of
**                           connected databases. This information should
**                           be transparent to business logic. Replaced
**                           database name string with Database
**                           information object. Required for non-local
**                           database connections.
** 006 ECF 20071012   @35590 Intermediate set of changes to support remote
**                           database connections more appropriately. This
**                           feature set is not yet fully implemented, but
**                           no existing functionality has been removed.
** 007 ECF 20071119   @35907 Completed implementation of remote database
**                           connections. Works with the net package to
**                           establish a virtual session with a remote
**                           server, to enable remote database access.
** 008 ECF 20071127   @36051 Added certAlias parameter to makeConnection()
**                           method. Passed through to SessionManager's
**                           connectVirtual() method. Currently, parameter
**                           is always null, indicating no requester-side
**                           validation of remote server's certificate
**                           will be performed by SecurityManager.
** 009 ECF 20080114   @36803 Fixed log output. Replaced incorrect format
**                           string.
** 010 ECF 20080118   @36858 Added logging to connect logic and connected
**                           check code.
** 011 ECF 20080213   @37069 Enable logical connections to locally managed
**                           databases. If a connection is attempted to a
**                           locally managed database, we short-circuit
**                           the logic which creates a virtual session and
**                           instead just manage the logical connection
**                           information. Fixed reconnect logic for
**                           pending disconnects.
** 012 SVL 20080806   @39313 Into disconnectImmediately() the session is
**                           terminated only if it is not used by the
**                           dirty share manager.
** 013 CA  20100223   @44686 When disconnecting from a remote database, 
**                           the session termination is done in a queue 
**                           (by using a global lock).
** 014 CA  20100329   @44784 Backed out H013 (@44686), as it was replaced 
**                           with a safer approach (the locking is done in 
**                           RouterSessionManager.terminateVirtual).
** 015 ECF 20120217          Added getPersistence(String) to allow the
**                           Persistence object associated with the given
**                           database to be used publicly.
** 016 OM  20121101          Implemented DBPARAM built-in function.
** 017 VMN 20130411          Added boolean parameter modal to recordOrShowError() method.
** 018 OM  20130507          Switch to int64 API function for obtaining desired java type.
** 019 OM  20130924          Added support for userid/password authorization to database.
**                           Switch to java.util.logging instead of Apache's logging API.
**                           Enriched DBPARAM result with -U/-P parameters.
*/

package com.goldencode.p2j.persist;

import java.net.InetSocketAddress;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.persist.dialect.P2JDialect;
import com.goldencode.p2j.persist.dirty.DirtyShareFactory;
import com.goldencode.p2j.security.ContextLocal;
import com.goldencode.p2j.util.*;
import org.hibernate.ScrollableResults;

/**
 * Manages information about connected databases in the current context.
 * This includes the existence of connections and their status, and the
 * various aliases and logical names assigned to physical databases.  This
 * information is managed to support converted Progress logic, which relies
 * upon the state of this information to make logic decisions.  It also is
 * used to determine which physical database is associated with each record
 * buffer and its backing {@link Persistence} service object.
 * <p>
 * A connection in the sense of this class is not a database-level connection,
 * such as a JDBC connection.  Database-level connections are managed by the
 * <code>Persistence</code> class transparently.  A connection to this class
 * is a logical construct inherited from converted Progress code.  Connecting
 * and disconnecting from a database at this level does not create or destroy
 * a JDBC-level connection resource.  However, for transient databases at
 * least, it may affect the loading and unloading of database configurations
 * at the {@link DatabaseManager} level.  That is, the first time any client
 * context connects to a transient database (i.e., one that is not always
 * loaded by the server by default), that database's configuration is loaded
 * temporarily by the server.  The last time a client context disconnects
 * from a transient database, that database's configuration is unloaded.
 * <p>
 * Connections register a physical database name to a logical database name,
 * which can be different or the same as the physical name.  A database may
 * not be connected more than once without first disconnecting.  Once
 * disconnected, it may be connected again.  Only one physical database may
 * be registered to a given logical name at a time.  Methods are provided to
 * connect and disconnect, as well as to query information about connected
 * databases, such as their physical and logical names, how many are
 * connected, their types and restrictions.
 * <p>
 * Any number of aliases may be assigned to a given logical name.  Aliases
 * may only be created for connected databases, though they survive database
 * disconnect.  They may be explicitly deleted, however.
 * <p>
 * Whenever a new record buffer scope is opened, it is registered with its
 * context's connection manager.  References from buffers to their backing,
 * physical databases are tracked, to enable the correct processing of
 * database disconnect requests.  Disconnects can only occur immediately if
 * all procedures which reference the target database are complete.  This
 * translates to a state where no record buffers reference the database.
 * Furthermore, even if no buffers reference a database, it may not be
 * disconnected if the active code point currently is within a transaction.
 * At the end of the transaction, all disconnects which were deferred are
 * processed.
 * <p>
 * A single instance of this class is created for each client context.  The
 * constructor is not called directly by client code;  rather, access is
 * provided through the static {@link #get} method.
 * <p>
 * Currently, the Progress semantic of database-level auto connect lists is
 * not supported.
 */
public final class ConnectionManager
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(ConnectionManager.class);
   
   /** Context local instance of this class */
   private static final ContextLocal<ConnectionManager> local =
      new ContextLocal<ConnectionManager>()
      {
         protected ConnectionManager initialValue()
         {
            return (new ConnectionManager());
         }
         
         protected void cleanup(ConnectionManager connMgr)
         {
            connMgr.sessionClosed();
         }
      };

   /**
    * Map with the userids for the currently connected databases.
    * This map must be kept in sync with the <code>connected</code> map of the databases.
    */
   private final Map<String, String> userIds = new LinkedHashMap<String, String>();

   /** 
    * Stores the userid used at connection time whether the database is connected using 
    * command-line (directory implementation) or programatically using <code>CONNECT</code>
    * with -U/-P parameters.
    * This list is not affected by later <code>SETUSERID</code> calls.
    */
   private final Map<String, String> connectTimeUserid = new LinkedHashMap<String, String>();
   
   /** 
    * Equivalent of -1 argument in CONNECT statement. At this moment it is not used
    * except for computing  DBPARAM function result. 
    * TODO: Future implementation will address this flag.
    */
   boolean singleClient = false;

   /**
    * Map of logical DB names to respective <code>Database</code> objects.
    * The values allow to extract the physical DB (root) names.
    */
   private final Map<String, Database> connected = new LinkedHashMap<String, Database>();
   
   /** Databases awaiting disconnect (first level) */
   private final Set<Database> pendingDisconn = new HashSet<Database>();
   
   /** Databases awaiting disconnect (second level) */
   private final Set<Database> pendingTxDisconn = new HashSet<Database>();
   
   /** Map of active aliases to logical database names */
   private final Map<String, String> aliases = new LinkedHashMap<String, String>();
   
   /** Map of buffer sets currently referencing databases */
   private final Map<Database, Set<RecordBuffer>> pdbReferences =
      new HashMap<Database, Set<RecordBuffer>>();
   
   /** Map of database info objects to network sessions */
   private final Map<Database, Session> sessions = new HashMap<Database, Session>();
   
   /**
    * Default constructor.  Registers with the transaction manager for master
    * and global finish notifications.  Auto-connects to all permanent
    * databases.
    */
   private ConnectionManager()
   {
      // Add all permanent databases to the map of connected databases, as
      // these are connected by default.  Logical names are the same as the physical names.
      Iterator<Database> iter = DatabaseManager.permanentConnections();
      while (iter.hasNext())
      {
         Database database = iter.next();
         
         // Do not include temp table dedicated database;  the fact that this
         // is a separate database instance is a P2J implementation detail
         // and should not bubble up to business logic.
         if (DatabaseManager.TEMP_TABLE_DB.equals(database))
         {
            continue;
         }
         
         connected.put(database.getName(), database);
         String userId = SecurityOps.BLANK_USER;
         // Query of _User table cannot be performed at this time. Calling Persistence.get()
         // now will create an infinite recursion as the context has not yet finished
         // its initialization.
         // However, during the authentication hook the userid-s will be updated to the correct
         // values using the values stored in directory tree or prompting the users otherwise.
         // Until then, the default blank user is used to keep userIds and connected maps in sync.
         userIds.put(database.getName(), userId);
      }
   }
   
   /**
    * Create a logical connection to a database using the specified
    * parameters.
    *
    * @param   args
    *          One or more connection parameters.  This may take one of
    *          several forms:
    *          <ul>
    *            <li>an ordered list of the following:
    *              <ol>
    *                <li>physical database name (required)
    *                <li>logical database name (optional)
    *                <li>P2J server host name/address (optional)
    *                <li>P2J server port number or service name (optional,
    *                    unless host name/address is supplied)
    *              </ol>
    *            <li>a string with the above listed components, prefaced by
    *                the following identifiers, respectively:
    *              <ol>
    *                <li>-db
    *                <li>-ld
    *                <li>-H
    *                <li>-S
    *              </ol>
    *              For example:  <code>-db mydb -ld logical_name -H
    *              192.168.1.12 -S 3333</code>.  After <code>-db</code>,
    *              parameters can be in any order, so long as each is
    *              preceded by the specified option identifier.
    *            <li>any combination of string fragments, which, if
    *                concatenated together in the order supplied, would create
    *                a string matching the requirements of the previous item.
    *          </ul>
    *
    * @throws  ErrorConditionException
    *          if any parameter is null or represents the unknown value;
    *          if any parameter is neither a text value nor an integral
    *          value;
    *          if there is an error registering the database with the
    *          database manager.
    */
   public static void connect(Object... args)
   {
      // Options will be parsed into individual components and stored in a
      // list for further processing.
      List<Object> options = new ArrayList<Object>();
      
      try
      {
         for (Object arg : args)
         {
            String text = null;
            Integer num = null;
            
            if (arg instanceof character)
            {
               character t = (character) arg;
               assertKnownValue(t, "Connect option");
               text = t.toStringMessage();
            }
            else if (arg instanceof String)
            {
               text = (String) arg;
            }
            else if (arg instanceof int64)
            {
               int64 i = (int64) arg;
               assertKnownValue(i, "Connect option");
               num = i.toJavaIntegerType();
            }
            else if (arg instanceof Integer)
            {
               num = (Integer) arg;
            }
            else
            {
               throw new IllegalArgumentException(
                  "Unexpected connect option:  " + arg);
            }
            
            if (text != null)
            {
               // Parse this string, which may represent a single connect
               // parameter, or multiple parameters delimited by whitespace.
               String[] parts = text.split(" ");
               for (String next : parts)
               {
                  // Add each separate parameter into the options list.
                  options.add(next.trim());
               }
            }
            else
            {
               // Add this non-text parameter into the options list.
               options.add(num);
            }
         }
         
         // Analyze all options and attempt connection.
         get().connectImpl(options);
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Request a disconnect from a logical database.  This does not actually
    * perform the disconnect immediately, but registers <code>ldbName</code>
    * for disconnect at the appropriate time:  once no buffers reference the
    * database, and at the end of the current transaction (if any).
    *
    * @param   ldbName
    *          Logical name of the database to disconnect.
    *
    * @throws  ErrorConditionException
    *          if the specified database has not been opened.
    */
   public static void disconnect(character ldbName)
   {
      try
      {
         assertKnownValue(ldbName, "LDB name");
         disconnect(ldbName.toStringMessage());
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Request a disconnect from a logical database.  This does not actually
    * perform the disconnect immediately, but registers <code>ldbName</code>
    * for disconnect at the appropriate time:  once no buffers reference the
    * database, and at the end of the current transaction (if any).
    *
    * @param   ldbName
    *          Logical name of the database to disconnect.
    *
    * @throws  ErrorConditionException
    *          if the specified database has not been opened.
    */
   public static void disconnect(String ldbName)
   {
      try
      {
         get().disconnectImpl(ldbName);
      }
      catch (PersistenceException exc)
      {
         ErrorManager.displayError(1009, 
                 "DISCONNECT:  Database " + ldbName + " is not open or unknown");
         // not a fatal error: recordOrThrowError(exc);
      }
   }
   
   /**
    * Tests whether a logical connection to a database exists in this context,
    * given a logical database name or alias.
    *
    * @param   name
    *          Logical database name or alias.
    *
    * @return  <code>true</code> if a logical connection exists;  else
    *          <code>false</code>.
    */
   public static logical connected(character name)
   {
      if (name.isUnknown())
      {
         return (new logical());
      }
      
      return connected(name.toStringMessage());
   }
   
   /**
    * Tests whether a logical connection to a database exists in this context,
    * given a logical database name or alias.
    *
    * @param   name
    *          Logical database name or alias.
    *
    * @return  <code>true</code> if a logical connection exists;  else
    *          <code>false</code>.
    */
   public static logical connected(String name)
   {
      return get().connectedImpl(name);
   }
   
   /**
    * Get the number of aliases currently defined in this context.
    *
    * @return  Number of aliases in this context.
    */
   public static integer numAliases()
   {
      return (new integer(get().aliases.size()));
   }
   
   /**
    * Get the number of databases currently connected in this context.
    *
    * @return  Number of logical connections in this context.
    */
   public static int getConnectedDbCount()
   {
      return get().connected.size();
   }
   
   /**
    * Get the number of databases currently connected in this context.
    *
    * @return  Number of logical connections in this context.
    */
   public static integer numDBs()
   {
      return new integer(getConnectedDbCount());
   }
   
   /**
    * Get the name of the connected database in this context.  If more than
    * one, return the first.
    *
    * @return  First or only connected database, or unknown value if none.
    */
   public static character dbName()
   {
      return pdbName(1);
   }

   /**
    * Implementation of the <code>DBPARAM()</code> Progress function.
    * Returns, as a character string, a comma-separated list of the parameters
    * used to connect to the database.
    *
    * <p>The sequence number of a database the Progress session is connected
    * to. For example, DBPARAM(1) returns information on the first database
    * the Progress session is connected to, DBPARAM(2) returns information
    * on the second database the Progress session is connected to, etc.
    * If you specify a sequence number that does not correspond to a database
    * the Progress session is connected to, the DBPARAM function returns the
    * unknown value (?).
    *
    * @param   index
    *          One-based index of the target database among all databases
    *          currently connected in this context.
    *
    * @return  A comma-separated list of the parameters used to connect to 
    *             the database
    */
   public static character dbParam(int index)
   {
      if (index == 0)
      {
         return new character();
      }

      return get().dbParamImpl(null, index);
   }

   /**
    * Implementation of the <code>DBPARAM()</code> Progress function.
    * Returns, as a character string, a comma-separated list of the parameters
    * used to connect to the database.
    *
    * <p>The sequence number of a database the Progress session is connected
    * to. For example, DBPARAM(1) returns information on the first database
    * the Progress session is connected to, DBPARAM(2) returns information
    * on the second database the Progress session is connected to, etc.
    * If you specify a sequence number that does not correspond to a database
    * the Progress session is connected to, the DBPARAM function returns the
    * unknown value (?).
    *
    * @param   index
    *          One-based index of the target database among all databases
    *          currently connected in this context.
    *
    * @return  A comma-separated list of the parameters used to connect to 
    *             the database
    */
   public static character dbParam(NumberType index)
   {
      if (index == null || index.isUnknown())
      {
         return new character();
      }

      return get().dbParamImpl(null, index.intValue());
   }

   /**
    * Implementation of the <code>DBPARAM()</code> Progress function.
    * Returns, as a character string, a comma-separated list of the parameters
    * used to connect to the database.
    *
    * <p>This forms of the <code>DBPARAM()</code function require a character
    * expression as a parameter. If the parameter is an alias or the logical
    * name of a connected database, then Progress returns the comma-separated
    * parameter list. Otherwise, it returns the unknown value (?).
    *
    * @param   nameOrAlias
    *          The logical-name or alias of a connected database
    *
    * @return  A comma-separated list of the parameters used to connect to 
    *             the database
    */
   public static character dbParam(String nameOrAlias)
   {
      if (nameOrAlias == null || nameOrAlias.equals(""))
      {
         return new character();
      }

      return get().dbParamImpl(nameOrAlias, 0);
   }

   /**
    * Implementation of the <code>DBPARAM()</code> Progress function.
    * Returns, as a character string, a comma-separated list of the parameters
    * used to connect to the database.
    *
    * <p>This forms of the <code>DBPARAM()</code function require a character
    * expression as a parameter. If the parameter is an alias or the logical
    * name of a connected database, then Progress returns the comma-separated
    * parameter list. Otherwise, it returns the unknown value (?).
    *
    * @param   nameOrAlias
    *          The logical-name or alias of a connected database
    *
    * @return  A comma-separated list of the parameters used to connect to 
    *             the database
    */
   public static character dbParam(character nameOrAlias)
   {
      if (nameOrAlias == null || nameOrAlias.isUnknown())
      {
         return new character();
      }

      return get().dbParamImpl(nameOrAlias.toStringMessage(), 0);
   }

   /**
    * Get a brief descriptor of the database type with the specified
    * connection index.
    *
    * @param   index
    *          One-based index of the target database among all databases
    *          currently connected in this context.
    *
    * @return  The string "PROGRESS" or unknown value if <code>index</code>
    *          represents the unknown value or is outside of the range of
    *          connected databases.
    */
   public static character dbType(NumberType index)
   {
      if (index.isUnknown())
      {
         return (new character());
      }
      
      return dbType(index.intValue());
   }
   
   /**
    * Get a brief descriptor of the database type with the specified
    * connection index.
    *
    * @param   index
    *          One-based index of the target database among all databases
    *          currently connected in this context.
    *
    * @return  The string "PROGRESS" or unknown value if <code>index</code>
    *          represents the unknown value or is outside of the range of
    *          connected databases.
    */
   public static character dbType(int index)
   {
      if (index < 1 || index > get().connected.size())
      {
         return (new character());
      }
      
      return (new character("PROGRESS"));
   }
   
   /**
    * Get a brief descriptor of the database type with the specified logical
    * name or alias.
    *
    * @param   name
    *          Logical name or alias of the target database.
    *
    * @return  The string "PROGRESS" or unknown value if <code>name</code>
    *          is not a recognized logical name or alias.
    */
   public static character dbType(character name)
   {
      if (name.isUnknown())
      {
         return (new character());
      }
      
      return dbType(name.toStringMessage());
   }
   
   /**
    * Get a brief descriptor of the database type with the specified logical
    * name or alias.
    *
    * @param   name
    *          Logical name or alias of the target database.
    *
    * @return  The string "PROGRESS" or unknown value if <code>name</code>
    *          is not a recognized logical name or alias.
    */
   public static character dbType(String name)
   {
      String ldb = get().getLDBName(name);
      if (ldb == null)
      {
         return (new character());
      }
      
      return (new character("PROGRESS"));
   }
   
   /**
    * Get a description of restrictions for the connected database at the
    * specified index.
    *
    * @param   index
    *          One-based index of the target database among all databases
    *          currently connected in this context.
    *
    * @return  The empty string, or unknown value if <code>index</code>
    *          represents the unknown value or is outside of the range of
    *          connected databases.
    */
   public static character dbRestrictions(NumberType index)
   {
      if (index.isUnknown())
      {
         return (new character());
      }
      
      return dbRestrictions(index.intValue(), (String) null);
   }
   
   /**
    * Get a description of restrictions for the connected database at the
    * specified index.
    *
    * @param   index
    *          One-based index of the target database among all databases
    *          currently connected in this context.
    *
    * @return  The empty string, or unknown value if <code>index</code> is
    *          invalid.
    */
   public static character dbRestrictions(int index)
   {
      return dbRestrictions(index, (String) null);
   }
   
   /**
    * Get a description of restrictions for the connected database at the
    * specified index.  Currently, <code>table</code> is ignored.
    *
    * @param   index
    *          One-based index of the target database among all databases
    *          currently connected in this context.
    * @param   table
    *          Name of a table within the target database.  Ignored.
    *
    * @return  The empty string, or unknown value if <code>index</code> is
    *          invalid.
    */
   public static character dbRestrictions(NumberType index, character table)
   {
      if (index.isUnknown())
      {
         return (new character());
      }
      
      return dbRestrictions(index.intValue(), (String) null);
   }
   
   /**
    * Get a description of restrictions for the connected database at the
    * specified index.  Currently, <code>table</code> is ignored.
    *
    * @param   index
    *          One-based index of the target database among all databases
    *          currently connected in this context.
    * @param   table
    *          Name of a table within the target database.  Ignored.
    *
    * @return  The empty string, or unknown value if <code>index</code> is
    *          invalid.
    */
   public static character dbRestrictions(NumberType index, String table)
   {
      if (index.isUnknown())
      {
         return (new character());
      }
      
      return dbRestrictions(index.intValue(), table);
   }
   
   /**
    * Get a description of restrictions for the connected database at the
    * specified index.  Currently, <code>table</code> is ignored.
    *
    * @param   index
    *          One-based index of the target database among all databases
    *          currently connected in this context.
    * @param   table
    *          Name of a table within the target database.  Ignored.
    *
    * @return  The empty string, or unknown value if <code>index</code> is
    *          invalid.
    */
   public static character dbRestrictions(int index, character table)
   {
      return dbRestrictions(index, (String) null);
   }
   
   /**
    * Get a description of restrictions for the connected database at the
    * specified index.  Currently, <code>table</code> is ignored.
    *
    * @param   index
    *          One-based index of the target database among all databases
    *          currently connected in this context.
    * @param   table
    *          Name of a table within the target database.  Ignored.
    *
    * @return  The empty string, or unknown value if <code>index</code> is
    *          invalid.
    */
   public static character dbRestrictions(int index, String table)
   {
      if (index < 1 || index > get().connected.size())
      {
         return (new character());
      }
      
      return (new character(""));
   }
   
   /**
    * Get a description of restrictions for the connected database with the
    * specified logical name or alias.
    *
    * @param   name
    *          Logical name or alias of the target database.
    *
    * @return  The empty string, or unknown value if <code>name</code>
    *          is not a recognized logical name or alias.
    */
   public static character dbRestrictions(character name)
   {
      return dbRestrictions(name, (String) null);
   }
   
   /**
    * Get a description of restrictions for the connected database with the
    * specified logical name or alias.
    *
    * @param   name
    *          Logical name or alias of the target database.
    *
    * @return  The empty string, or unknown value if <code>name</code>
    *          is not a recognized logical name or alias.
    */
   public static character dbRestrictions(String name)
   {
      return dbRestrictions(name, (String) null);
   }
   
   /**
    * Get a description of restrictions for the connected database with the
    * specified logical name or alias.  Currently, <code>table</code> is
    * ignored.
    *
    * @param   name
    *          Logical name or alias of the target database.
    * @param   table
    *          Name of a table within the target database.  Ignored.
    *
    * @return  The empty string, or unknown value if <code>name</code>
    *          is not a recognized logical name or alias.
    */
   public static character dbRestrictions(character name, character table)
   {
      if (name.isUnknown())
      {
         return (new character());
      }
      
      return dbRestrictions(name.toStringMessage(), table.toStringMessage());
   }
   
   /**
    * Get a description of restrictions for the connected database with the
    * specified logical name or alias.  Currently, <code>table</code> is
    * ignored.
    *
    * @param   name
    *          Logical name or alias of the target database.
    * @param   table
    *          Name of a table within the target database.  Ignored.
    *
    * @return  The empty string, or unknown value if <code>name</code>
    *          is not a recognized logical name or alias.
    */
   public static character dbRestrictions(character name, String table)
   {
      if (name.isUnknown())
      {
         return (new character());
      }
      
      return dbRestrictions(name.toStringMessage(), table);
   }
   
   /**
    * Get a description of restrictions for the connected database with the
    * specified logical name or alias.  Currently, <code>table</code> is
    * ignored.
    *
    * @param   name
    *          Logical name or alias of the target database.
    * @param   table
    *          Name of a table within the target database.  Ignored.
    *
    * @return  The empty string, or unknown value if <code>name</code>
    *          is not a recognized logical name or alias.
    */
   public static character dbRestrictions(String name, character table)
   {
      return dbRestrictions(name, (String) null);
   }
   
   /**
    * Get a description of restrictions for the connected database with the
    * specified logical name or alias.  Currently, <code>table</code> is
    * ignored.
    *
    * @param   name
    *          Logical name or alias of the target database.
    * @param   table
    *          Name of a table within the target database.  Ignored.
    *
    * @return  The empty string, or unknown value if <code>name</code>
    *          is not a recognized logical name or alias.
    */
   public static character dbRestrictions(String name, String table)
   {
      String ldb = get().getLDBName(name);
      if (ldb == null)
      {
         return (new character());
      }
      
      return (new character(""));
   }
   
   /**
    * Create an alias for a logical database name within the current context.
    * The alias may not conflict with an existing logical database name
    * (except for that of the database which it aliases).  The alias will be
    * created regardless of whether the target database currently is
    * connected;  however, an error will be raised (or silently registered)
    * in the event it is not, after the alias has been created.  The alias
    * persists until explicitly deleted with one of the variants of the
    * <code>deleteAlias</code> method, or until the current session ends.
    *
    * @param   alias
    *          Alias to be created.
    * @param   ldbName
    *          Logical name of the database to which the alias is assigned.
    *
    * @throws  ErrorConditionException
    *          if silent error mode is disabled and any of the following:
    *          <ul>
    *            <li><code>alias</code> represents the unknown value;
    *            <li><code>ldbName</code> represents the unknown value;
    *            <li>the target database is not connected;
    *            <li>the alias conflicts with an existing logical database
    *                name, other than <code>ldbName</code>.
    *          </ul>
    */
   public static void createAlias(character alias, character ldbName)
   {
      try
      {
         assertKnownValue(alias, "Alias");
         assertKnownValue(ldbName, "LDB name");
         createAlias(alias.toStringMessage(), ldbName.toStringMessage());
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Create an alias for a logical database name within the current context.
    * The alias may not conflict with an existing logical database name
    * (except for that of the database which it aliases).  The alias will be
    * created regardless of whether the target database currently is
    * connected;  however, an error will be raised (or silently registered)
    * in the event it is not, after the alias has been created.  The alias
    * persists until explicitly deleted with one of the variants of the
    * <code>deleteAlias</code> method, or until the current session ends.
    *
    * @param   alias
    *          Alias to be created.
    * @param   ldbName
    *          Logical name of the database to which the alias is assigned.
    *
    * @throws  ErrorConditionException
    *          if silent error mode is disabled and any of the following:
    *          <ul>
    *            <li><code>ldbName</code> represents the unknown value;
    *            <li>the target database is not connected;
    *            <li>the alias conflicts with an existing logical database
    *                name, other than <code>ldbName</code>.
    *          </ul>
    */
   public static void createAlias(String alias, character ldbName)
   {
      try
      {
         assertKnownValue(ldbName, "LDB name");
         createAlias(alias, ldbName.toStringMessage());
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Create an alias for a logical database name within the current context.
    * The alias may not conflict with an existing logical database name
    * (except for that of the database which it aliases).  The alias will be
    * created regardless of whether the target database currently is
    * connected;  however, an error will be raised (or silently registered)
    * in the event it is not, after the alias has been created.  The alias
    * persists until explicitly deleted with one of the variants of the
    * <code>deleteAlias</code> method, or until the current session ends.
    *
    * @param   alias
    *          Alias to be created.
    * @param   ldbName
    *          Logical name of the database to which the alias is assigned.
    *
    * @throws  ErrorConditionException
    *          if silent error mode is disabled and any of the following:
    *          <ul>
    *            <li><code>alias</code> represents the unknown value;
    *            <li>the target database is not connected;
    *            <li>the alias conflicts with an existing logical database
    *                name, other than <code>ldbName</code>.
    *          </ul>
    */
   public static void createAlias(character alias, String ldbName)
   {
      try
      {
         assertKnownValue(alias, "Alias");
         createAlias(alias.toStringMessage(), ldbName);
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Create an alias for a logical database name within the current context.
    * The alias may not conflict with an existing logical database name
    * (except for that of the database which it aliases).  The alias will be
    * created regardless of whether the target database currently is
    * connected;  however, an error will be raised (or silently registered)
    * in the event it is not, after the alias has been created.  The alias
    * persists until explicitly deleted with one of the variants of the
    * <code>deleteAlias</code> method, or until the current session ends.
    *
    * @param   alias
    *          Alias to be created.
    * @param   ldbName
    *          Logical name of the database to which the alias is assigned.
    *
    * @throws  ErrorConditionException
    *          if silent error mode is disabled and any of the following:
    *          <ul>
    *            <li>the target database is not connected;
    *            <li>the alias conflicts with an existing logical database
    *                name, other than <code>ldbName</code>.
    *          </ul>
    */
   public static void createAlias(String alias, String ldbName)
   {
      try
      {
         get().createAliasImpl(alias, ldbName);
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Delete the specified alias immediately.
    *
    * @param   alias
    *          Alias to be deleted.
    */
   public static void deleteAlias(character alias)
   {
      if (alias.isUnknown())
      {
         return;
      }
      
      deleteAlias(alias.toStringMessage());
   }
   
   /**
    * Delete the specified alias immediately.
    *
    * @param   alias
    *          Alias to be deleted.
    */
   public static void deleteAlias(String alias)
   {
      get().deleteAliasImpl(alias);
   }
   
   /**
    * Retrieve the alias at the specified index in the current context.
    *
    * @param   index
    *          One-based index of the alias (aliases are stored in the order
    *          in which they were created).
    *
    * @return  The target alias, or unknown value if <code>index</code> is
    *          invalid.
    */
   public static character alias(NumberType index)
   {
      if (index.isUnknown())
      {
         return (new character());
      }
      
      return alias(index.intValue());
   }
   
   /**
    * Retrieve the alias at the specified index in the current context.
    *
    * @param   index
    *          One-based index of the alias (aliases are stored in the order
    *          in which they were created).
    *
    * @return  The target alias, or unknown value if <code>index</code> is
    *          invalid.
    */
   public static character alias(int index)
   {
      return get().aliasImpl(index);
   }
   
   /**
    * Get the logical database name for the connected database at the given
    * index.
    *
    * @param   index
    *          One-based index of the target database among all databases
    *          currently connected in this context.
    *
    * @return  The logical name, or unknown value if <code>index</code> is
    *          invalid.
    */
   public static character ldbName(NumberType index)
   {
      if (index.isUnknown())
      {
         return (new character());
      }
      
      return ldbName(index.intValue());
   }
   
   /**
    * Get the logical database name for the connected database at the given
    * index.
    *
    * @param   index
    *          One-based index of the target database among all databases
    *          currently connected in this context.
    *
    * @return  The logical name, or unknown value if <code>index</code> is
    *          invalid.
    */
   public static character ldbName(int index)
   {
      return get().ldbNameImpl(null, index);
   }
   
   /**
    * Get the logical database name for the connected database with the given
    * logical name or alias.
    *
    * @param   name
    *          Logical name or alias of the target database.
    *
    * @return  The logical name, or unknown value if <code>name</code> is
    *          invalid.
    */
   public static character ldbName(character name)
   {
      if (name.isUnknown())
      {
         return (new character());
      }
      
      return ldbName(name.toStringMessage());
   }
   
   /**
    * Get the logical database name for the connected database with the given
    * logical name or alias.
    *
    * @param   name
    *          Logical name or alias of the target database.
    *
    * @return  The logical name, or unknown value if <code>name</code> is
    *          invalid.
    */
   public static character ldbName(String name)
   {
      return get().ldbNameImpl(name, 0);
   }
   
   /**
    * Get the logical database name for the record buffer associated with the
    * given DMO.
    *
    * @param   dmo
    *          Data model object returned by a previous call to {@link
    *          RecordBuffer#define}.
    *
    * @return  The logical database name for the DMO.
    */
   public static character ldbName(Object dmo)
   {
      RecordBuffer buffer = ((BufferReference) dmo).buffer();
      
      return (new character(buffer.getLogicalDatabase()));
   }
   
   /**
    * Get the physical database name for the connected database at the given
    * index.
    *
    * @param   index
    *          One-based index of the target database among all databases
    *          currently connected in this context.
    *
    * @return  The physical name, or unknown value if <code>index</code> is
    *          invalid.
    */
   public static character pdbName(NumberType index)
   {
      if (index.isUnknown())
      {
         return (new character());
      }
      
      return pdbName(index.intValue());
   }
   
   /**
    * Get the physical database name for the connected database at the given
    * index.
    *
    * @param   index
    *          One-based index of the target database among all databases
    *          currently connected in this context.
    *
    * @return  The physical name, or unknown value if <code>index</code> is
    *          invalid.
    */
   public static character pdbName(int index)
   {
      return get().pdbNameImpl(null, index);
   }
   
   /**
    * Get the physical database name for the connected database with the
    * given logical name or alias.
    *
    * @param   name
    *          Logical name or alias of the target database.
    *
    * @return  The physical name, or unknown value if <code>name</code> is
    *          invalid.
    */
   public static character pdbName(character name)
   {
      if (name.isUnknown())
      {
         return (new character());
      }
      
      return pdbName(name.toStringMessage());
   }
   
   /**
    * Get the physical database name for the connected database with the
    * given logical name or alias.
    *
    * @param   name
    *          Logical name or alias of the target database.
    *
    * @return  The physical name, or unknown value if <code>name</code> is
    *          invalid.
    */
   public static character pdbName(String name)
   {
      return get().pdbNameImpl(name, 0);
   }
   
   /**
    * Get the logical database name for the connected database at the given
    * index.
    * <p>
    * The current implementation delegates to {@link #ldbName(NumberType)}.
    *
    * @param   index
    *          One-based index of the target database among all databases
    *          currently connected in this context.
    *
    * @return  The logical name, or unknown value if <code>index</code> is
    *          invalid.
    */
   public static character sdbName(NumberType index)
   {
      if (index.isUnknown())
      {
         return (new character());
      }
      
      return sdbName(index.intValue());
   }
   
   /**
    * Get the logical database name for the connected database at the given
    * index.
    * <p>
    * The current implementation delegates to {@link #ldbName(int)}.
    *
    * @param   index
    *          One-based index of the target database among all databases
    *          currently connected in this context.
    *
    * @return  The logical name, or unknown value if <code>index</code> is
    *          invalid.
    */
   public static character sdbName(int index)
   {
      return ldbName(index);
   }
   
   /**
    * Get the logical database name for the connected database with the given
    * logical name or alias.
    * <p>
    * The current implementation delegates to {@link #ldbName(character)}.
    *
    * @param   name
    *          Logical name or alias of the target database.
    *
    * @return  The logical name, or unknown value if <code>name</code> is
    *          invalid.
    */
   public static character sdbName(character name)
   {
      if (name.isUnknown())
      {
         return (new character());
      }
      
      return sdbName(name.toStringMessage());
   }
   
   /**
    * Get the logical database name for the connected database with the given
    * logical name or alias.
    * <p>
    * The current implementation delegates to {@link #ldbName(String)}.
    *
    * @param   name
    *          Logical name or alias of the target database.
    *
    * @return  The logical name, or unknown value if <code>name</code> is
    *          invalid.
    */
   public static character sdbName(String name)
   {
      return ldbName(name);
   }
   
   /**
    * Retrieve the network session which manages the connection to the remote
    * server which is authoritative for the specified database.
    * 
    * @param   database
    *          Database to which the target session is mapped.
    * 
    * @return  Network session for the given database, or <code>null</code>
    *          if there is no current connection.
    */
   public static Session getSession(Database database)
   {
      return get().sessions.get(database);
   }
   
   /**
    * Get the {@link Persistence} instance associated with the given database
    * name, which may be a physical or logical name, or an alias.
    *
    * @param   name
    *          Physical or logical database name, or alias.
    *          
    * @return  Persistence object associated with the given database, which
    *          may be locally or remotely connected.
    */
   public static Persistence getPersistence(String name)
   {
      Database database = get().getDatabase(name);
      
      return PersistenceFactory.getInstance(database);
   }

   /**
    * Returns the current id for current that the user is authenticated to a database
    * identified by its logical name.
    *
    * @param   ldbName
    *          The logical name of the database for which the userid is requested.
    *
    * @return  current userid
    */
   public static String getCurrentUserid(String ldbName)
   {
      return get().getCurrentUseridImpl(ldbName);
   }
   
   /**
    * Authenticates the user for specified DB connection. Checks if the user login account match
    * the corresponding record of the _User table of the database.  If the specified database
    * is not connected, an error is issued.
    *
    * The user/password comparing operation is done as 4GL:
    * <ul>
    *    <li><strong>user</strong>: is case-insensitive, trailing spaces are ignored
    *             (in upper(rtrim()) mode)
    *    <li><strong>password</strong>: must be an exact match: after applying <code>ENCODE</code>
    *             to password string it is compared to existing record from database in case
    *             sensitive, taking into consideration trailing spaces
    * </ul>
    *
    * @param    userid
    *           The name of the user to set as UserID.
    * @param    password
    *           The users password.
    * @param    dbname
    *           Logical database name.
    *
    * @return   The <code>true</code> in case of valid user match <code>false</code> otherwise.
    */
   public static boolean authenticate(String userid, String password, String dbname)
   {
      ConnectionManager ctxCm = get();

      // validation: dbname MUST be the logical name of a connected database
      character ldbName = ctxCm.ldbNameImpl(dbname, 0);
      if (dbname == null      ||
          dbname.isEmpty()    ||
          ldbName.isUnknown() ||
          !dbname.equalsIgnoreCase(ldbName.toStringMessage()))
      {
         // the 'dbname' is not a valid connected database logical name nor an alias
         ErrorManager.recordOrShowError(
               1072,
               "SETUSERID of database " + dbname +
                     " requires that database to be connected", false);
         // at runtime, we get a false value (not unknown, not error)
         return false;
      }

      String encPassword = SecurityOps.encode(password).toStringMessage();
      Persistence persistence = PersistenceFactory.getInstance(ctxCm.getDatabase(dbname));

      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String pccUserid = dialect.getProcessedCharacterColumnName(
               MetadataSecurityOps.META_FIELD_USERID, true);
      String pccPassword = MetadataSecurityOps.META_FIELD_PASSWORD;
               // SETUSERID() is strict when comparing passwords

      String query =  "SELECT COUNT(*) FROM " + MetadataSecurityOps.META_TABLE_USER + " " +
                           "WHERE " + pccUserid + " = upper(rtrim(? , ' \t\n\r')) " +
                           "AND " + pccPassword + " = ? ";
      try
      {
         ScrollableResults rs = persistence.executeSQLQuery(query,
                                                            new String[] { userid, encPassword });
         if (rs == null || !rs.next())
         {
            // this should not happen anyway
            return false;
         }

         long count = ((java.math.BigInteger) rs.get(0)).longValue();
         if (count != 1)
         {
            // no match (if zero), if 2 or more then the _user table is broken
            return false;
         }
      }
      catch (PersistenceException pe)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, "setUserId failed accessing database (SQL query)", pe);
         }
         return false;
      }

      // if we reached this point then the user/password combination is OK
      // (the userid may be in camel-case or have trailing spaces)
      return true;
   }

   /**
    * Check the access level for a database based on metadata of that database.
    * This will query if _User table and return <code>SecurityOps.FREE_ACCESS</code> if
    * the meta table is not populated and <code>SecurityOps.BLANK_ALLOWED</code> otherwise.
    *
    * @param   ldbname
    *          The logical database name to query.
    *
    * @return  The access level to specified database.
    */
   public static int getAuthLevel(String ldbname)
   {
      String query =  "SELECT COUNT(*) FROM " + MetadataSecurityOps.META_TABLE_USER;
      try
      {
         ScrollableResults rs = getPersistence(ldbname).executeSQLQuery(query);
         if (rs != null && rs.next())
         {
            long count = ((java.math.BigInteger) rs.get(0)).longValue();
            return count == 0 ? SecurityOps.FREE_ACCESS : SecurityOps.BLANK_ALLOWED;
         }
      }
      catch (PersistenceException pe)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            LOG.log(Level.WARNING, "setUserId failed accessing database (SQL query)", pe);
         }
      }

      return SecurityOps.FREE_ACCESS;
   }

   /**
    * Set the currently authenticated userid for a database.
    * This method assumes that the userid has already been authenticated in this context using
    * <code>authenticate()</code> method.
    * If <code>byConnect</code> then the userid is stored and  will be reported by
    * <code>dbparam</code> function with -U tag.
    *
    * @param   ldbName
    *          Logical database name.
    * @param   userid
    *          The name of the user to set as UserID.
    * @param   byConnect
    *          <code>true</code>if the authentication was done from a connect (at startup time
    *          or later) or <code>false</code> if the authentication was done using the
    *          <code>setuserid</code> function.
    */
   public static void setAuthenticatedUserid(String ldbName, String userid, boolean byConnect)
   {
      get().setCurrentUseridImpl(ldbName, userid);
      if (byConnect)
      {
         get().setConnectUserImpl(ldbName);
      }
   }

   /**
    * Retrieve the <code>ConnectionManager</code> instance for the local context.
    *
    * @return  Context local connection manager.
    */
   static ConnectionManager get()
   {
      return local.get();
   }
   
   /**
    * Assert that <code>data</code> does not represent the unknown value.
    * If the assertion fails, an exception is thrown.
    *
    * @param   data
    *          The data value to be checked.
    * @param   description
    *          A brief description of <code>data</code>'s purpose;  this will
    *          be used in an error message if the assertion fails.
    *
    * @throws  PersistenceException
    *          if the assertion fails.
    */
   private static void assertKnownValue(BaseDataType data, String description)
   throws PersistenceException
   {
      if (data.isUnknown())
      {
         // TODO:  use proper error number/message.
         throw new PersistenceException(
            description + " cannot be unknown",
            -1);
      }
   }
   
   /**
    * Iterate over the specified collection, returning the element at index
    * <code>index</code>.
    *
    * @param   coll
    *          Collection over which to iterate.
    * @param   index
    *          The <b>1-based</b> index of the target element in the
    *          collection.
    *
    * @return  The element at the specified index, or <code>null</code> if
    *          <code>index</code> was never reached.
    */
   private static Object iterateToIndex(Collection coll, int index)
   {
      Iterator iter = coll.iterator();
      Object element = null;
      
      for (int i = 1; iter.hasNext(); i++)
      {
         Object next = iter.next();
         if (i == index)
         {
            element = next;
            break;
         }
      }
      
      return element;
   }
   
   /**
    * Normalize the given filename by extracting the root name (removing the
    * file path and the filename extension).
    *
    * @param   name
    *          File name.
    *
    * @return  "Normalized" root file name.
    */
   private static String normalizeFilename(String name)
   {
      // Extract root name from PDB name.
      int pos = name.lastIndexOf(".");
      if (pos >= 0)
      {
         name = name.substring(0, pos);
      }
      pos = name.lastIndexOf(System.getProperty("file.separator"));
      if (pos >= 0)
      {
         name = name.substring(pos + 1);
      }
      
      return name;
   }

   /**
    * This method is invoked after a master transaction commit/rollback;  any
    * databases awaiting the end of the current transaction to disconnect are
    * now deregistered with the database manager.
    */
   void transactionEnded()
   {
      disconnectImmediately(pendingTxDisconn);
   }
   
   /**
    * Register that a reference from a record buffer to its database has
    * become active.  We will be notified when this reference is no longer
    * active, which is used to resolve pending disconnect requests.
    *
    * @param   buffer
    *          Record buffer whose scope has been opened.
    *
    * @see     #closeBufferScope
    * @see     #disconnect
    */
   void openBufferScope(RecordBuffer buffer)
   {
      // Add buffer to set of those which reference a physical database.
      Database pdb = buffer.getDatabase();
      Set<RecordBuffer> set = pdbReferences.get(pdb);
      if (set == null)
      {
         set = new HashSet<RecordBuffer>();
         pdbReferences.put(pdb, set);
      }
      
      set.add(buffer);
   }
   
   /**
    * Called by a record buffer to notify that its scope has been closed,
    * indicating that it no longer has an active reference to its database.
    * This may result in the current context disconnecting from the database,
    * if and only if:
    * <ol>
    *   <li>there is a (first level) disconnect request pending for the
    *       database;  AND
    *   <li>no record buffers currently have active references to it;  AND
    *   <li>no transaction currently is active.
    * </ol>
    * <p>
    * If all three conditions are met, the database is disconnected
    * immediately, and removed from the first level pending disconnect list.
    * If the first two conditions are met, but the third is not, the database
    * is moved from the first level to the second level disconnect list,
    * where it will await the end of the transaction for disconnect.
    *
    * @param   buffer
    *          Record buffer whose scope has been closed.
    *
    * @see     #openBufferScope
    * @see     #disconnect
    */
   void closeBufferScope(RecordBuffer buffer)
   {
      // Nothing to do if this is not the outermost open scope for this
      // buffer.
      if (buffer.getOpenScopeCount() > 0)
      {
         return;
      }
      
      // Remove buffer from set of those which reference a physical database.
      Database pdb = buffer.getDatabase();
      Set<RecordBuffer> set = pdbReferences.get(pdb);
      set.remove(buffer);
      
      // If the set is now empty, we are not in a transaction, AND this
      // database is pending disconnect, deregister it with the database
      // manager.
      if (set.isEmpty())
      {
         pdbReferences.remove(pdb);
         if (pendingDisconn.contains(pdb))
         {
            if (!TransactionManager.isTransaction())
            {
               // Database can be deregistered immediately.
               DatabaseManager.deregisterDatabase(pdb);
            }
            else
            {
               // Database cannot be deregistered until transaction ends.
               pendingTxDisconn.add(pdb);
            }
            
            // Either way, remove it from the first level pending disconnect
            // set.
            pendingDisconn.remove(pdb);
         }
      }
   }
   
   /**
    * Get the logical database name for the given logical database name or
    * alias.  The database does not necessarily have to be connected, if
    * <code>name</code> represents an alias.
    *
    * @param   name
    *          Logical database name or alias.
    *
    * @return  Logical database name or <code>null</code> if <code>name</code>
    *          is not recognized.
    */
   String getLDBName(String name)
   {
      name = name.toLowerCase();
      
      // Look up name as an alias.
      String ldbName = aliases.get(name);
      if (ldbName != null)
      {
         // name is an alias.
         return ldbName;
      }
      else if (connected.containsKey(name))
      {
         // name is the LDB name.
         return name;
      }
      
      // name is unrecognized.
      return null;
   }
   
   /**
    * Get the physical database name for the given logical database name or
    * alias.  The database must be connected.  The resulting name may
    * optionally be normalized, which excludes any filename path and
    * extension.
    *
    * @param   name
    *          Logical database name or alias.
    * @param   normalize
    *          <code>true</code> to return the normalized name;  else
    *          <code>false</code> to return the unaltered (possibly fully
    *          qualified name), as provided by the caller of the {@link
    *          #connect} method.
    *
    * @return  Physical database name or <code>null</code> if
    *          <code>name</code> is not recognized.
    */
   String getPDBName(String name, boolean normalize)
   {
      String pdb = getDatabase(name).getName();
      if (normalize)
      {
         pdb = normalizeFilename(pdb);
      }
      
      return pdb;
   }
   
   /**
    * Get the database information object for the given database name, which
    * may be a logical name, physical name, or alias.  It is assumed that the
    * database currently is connected.
    * 
    * @param   name
    *          Logical name, physical name, or alias for database.
    *          
    * @return  Database information object associated with <code>name</code>,
    *          or <code>null</code> if no such database currently is
    *          connected.
    */
   Database getDatabase(String name)
   {
      // Get LDB name, in case name is an alias.
      String ldbName = getLDBName(name);
      
      return connected.get(ldbName);
   }

   /**
    * Returns the current id used by the user to authenticated to a database
    * identified by its logical name.
    *
    * @param   ldbName
    *          The logical name of the database for which the userid is requested.
    *
    * @return  current userid
    */
   private String getCurrentUseridImpl(String ldbName)
   {
      if (userIds.containsKey(ldbName))
      {
         return userIds.get(ldbName);
      }
      return "";
   }

   /**
    * Sets the currently authenticated userid for a database.
    * The only authority to set this is an instance of <code>MetadataSecurityOps</code> and only
    * after checking the user/password combination against the _User meta-table.
    *
    * Notes:
    *  * The CONNECT statement with -U and -P parameters will have to call SecurityOps to verify
    *       if the user/password combination if correct.
    *       (At this moment the ConnectionManager does not support these parameters, yet).
    *  * The command-line parameters should be investigated.
    *
    * @param   ldbName
    *          The logical database name for which the userid is set/updated.
    * @param   newUserId
    *          The newly newUserId for this database context.
    */
   private void setCurrentUseridImpl(String ldbName, String newUserId)
   {
      userIds.put(ldbName, newUserId);
   }
   
   /**
    * Mark the current user for the database as the connection parameter.
    * The user currently authenticated will be reported by <code>dbparam</code> function with -U  
    * tag.
    *
    * @param   ldbName
    *          The logical database name for which the userid is maked as connection parameter.
    */
   private void setConnectUserImpl(String ldbName)
   {
      String userId = userIds.get(ldbName);
      if (userId != null)
      {
         connectTimeUserid.put(ldbName, userId);
      }
   }

   /**
    * Perform end of session cleanup;  disconnects all connected databases
    * without regard to transaction state or buffer reference status.  This
    * ensures <code>DatabaseManager</code>'s transient database reference
    * counts are not corrupted when client sessions end.
    */
   private void sessionClosed()
   {
      // waistband
      Set<Database> all = new HashSet<Database>(connected.values());
      // belt
      all.addAll(pendingTxDisconn);
      // suspenders
      all.addAll(pendingDisconn);
      
      // Deregister all connected databases.
      disconnectImmediately(all);
   }
   
   /**
    * Disconnect all databases in the specified set by deregistering them
    * with the <code>DatabaseManager</code> and removing them from the set.
    * Network sessions associated with the databases in the set are
    * terminated accordingly.
    *
    * @param   pending
    *          Set of databases to disconnect.
    */
   private void disconnectImmediately(Set<Database> pending)
   {
      Iterator<Database> iter = pending.iterator();
      while (iter.hasNext())
      {
         Database database = iter.next();
         disconnectImmediately(database);
         iter.remove();
      }
   }
   
   /**
    * Disconnect the specified database by deregistering it with the
    * <code>DatabaseManager</code> and terminating the associated network
    * session.
    * 
    * @param   database
    *          Database to disconnect.
    */
   private void disconnectImmediately(Database database)
   {
      if (!database.isLocal())
      {
         DatabaseManager.deregisterDatabase(database);
      }
      
      Session session = sessions.remove(database);
      if (session != null)
      {
         // Close the session only if is not used by the dirty share manager
         // for this database.
         if (session.isRunning() && !DirtyShareFactory.isUsedByManager(database, session))
         {
            session.terminate();
         }
      }
      
      if (LOG.isLoggable(Level.INFO))
      {
         String ctx = Utils.describeContext();
         // database is not null
         String msg = String.format("{%s} Database '%s' disconnected", ctx, database.toString());
         LOG.log(Level.INFO, msg);
         
         if (LOG.isLoggable(Level.FINE))
         {
            msg = String.format("{%s} connected:  %s", ctx, connected.toString());
            LOG.log(Level.FINE, msg);
            msg = String.format("{%s} pendingDisconn:  %s", ctx, pendingDisconn.toString());
            LOG.log(Level.FINE, msg);
            msg = String.format("{%s} pendingTxDisconn:  %s", ctx, pendingTxDisconn.toString());
            LOG.log(Level.FINE, msg);
         }
      }
   }
   
   /**
    * Get the logical name for a database, given an alias (which may be the
    * logical name itself), or its index among all connected databases in
    * this context.
    *
    * @param   name
    *          Logical database name or alias.  May be <code>null</code>,
    *          but only if <code>index</code> is non-<code>null</code>.
    * @param   index
    *          Index of the database among connected databases.  May be
    *          <code>null</code>, but only if <code>name</code> is
    *          non-<code>null</code>.  This is a one-based index.
    *
    * @return  Logical database name, or unknown value if <code>name</code>
    *          is not recognized.
    */
   private character ldbNameImpl(String name, int index)
   {
      String ldbName = null;
      if (name == null)
      {
         ldbName = (String) iterateToIndex(connected.keySet(), index);
      }
      else
      {
         ldbName = getLDBName(name);
      }
      
      if (ldbName != null)
      {
         return (new character(ldbName));
      }
      
      return (new character());
   }
   
   /**
    * Get the physical name for a database, given its logical name or alias,
    * or its index among all connected databases in this context.
    *
    * @param   name
    *          Logical database name or alias.  May be <code>null</code>,
    *          but only if <code>index</code> is non-<code>null</code>.
    * @param   index
    *          One-based index of the database among connected databases.
    *
    * @return  Physical database name, or unknown value if <code>name</code>
    *          is not recognized.
    */
   private character pdbNameImpl(String name, int index)
   {
      String key = null;
      if (name == null)
      {
         key = (String) iterateToIndex(connected.keySet(), index);
      }
      else
      {
         key = name.toLowerCase();
      }
      
      String pdbName = getPDBName(key, false);
      if (pdbName != null)
      {
         return (new character(pdbName));
      }
      
      return (new character());
   }
   
   /**
    * Create a logical connection to a database using the specified
    * parameters.
    *
    * @param   args
    *          A list of one or more connection parameters.  This may take one
    *          of several forms:
    *          <ul>
    *            <li>an ordered list of the following:
    *              <ol>
    *                <li>physical database name (required)
    *                <li>logical database name (optional)
    *                <li>P2J server host name/address (optional)
    *                <li>P2J server port number or service name (optional,
    *                    unless host name/address is supplied)
    *              </ol>
    *            </li>
    *            <li>the following parameter pairs:
    *              <ol>
    *                <li>physical database name key/value pair...
    *                  <ul>
    *                    <li>"-db" (optional identifier for physical database
    *                        name option to follow)
    *                    <li>physical database name (required)
    *                  </ul>
    *                </li>
    *                <li>...followed by an unordered list of the following
    *                    options.  Note that the order of these parameter
    *                    key/value pairs is not important;  however, the
    *                    pairing of key string (e.g., "-ld"), with the
    *                    associated value must be preserved:
    *                  <ul>
    *                    <li>"-ld" (required only if next element is provided)
    *                    <li>logical database name (optional)
    *                    <li>"-H" (required only if next element is provided)
    *                    <li>P2J server host name/address (optional)
    *                    <li>"-S" (required only if next element is provided)
    *                    <li>P2J server port number or service name (optional,
    *                        unless host name/address is supplied)
    *                  </ul>
    *                </li>
    *              </ol>
    *            </li>
    *          </ul>
    *
    * @throws  PersistenceException
    *          if any parameter is null;
    *          if any parameter is neither a text value nor an integral
    *          value;
    *          if there is an error registering the database with the
    *          database manager.
    */
   private void connectImpl(List<Object> args)
   throws PersistenceException
   {
      String flag = null;
      String pdb = null;
      String ldb = null;
      String host = null;
      String service = null;
      String user = null;
      String password = null;
      Integer port = null;
      
      // Loop through the parameters and divine their meanings.
      for (Object arg : args)
      {
         String text = null;
         Integer num = null;
         
         if (arg instanceof String)
         {
            text = (String) arg;
         }
         else if (arg instanceof Integer)
         {
            num = (Integer) arg;
         }
         else if (arg == null)
         {
            throw new PersistenceException("Invalid (null) connect argument");
         }
         
         if (text != null)
         {
            // Detect option keys.
            if (text.startsWith("-"))
            {
               if (flag != null)
               {
                  // TODO: both error messages are printed on P4GL
                  ErrorManager.throwError(1403,
                        "You have not supplied a parameter for argument " + flag, null);
                  ErrorManager.throwError(325, "Unable to execute CONNECT statement", null);
               }
               
               if (text.equals("-1"))
               {
                  singleClient = true;
               }
               else
               {
                  flag = text;
               }
               
               continue;
            }
            
            // Physical database name.
            if ("-db".equals(flag) || (flag == null && pdb == null))
            {
               pdb = text;
               flag = null;
               
               continue;
            }
            
            // Logical database name.
            if ("-ld".equals(flag) || (flag == null && ldb == null))
            {
               ldb = text;
               flag = null;
               
               continue;
            }
            
            // Host name/address.
            if ("-H".equals(flag) || (flag == null && host == null))
            {
               host = text;
               flag = null;
               
               continue;
            }

            // Authentication: userid
            if ("-U".equals(flag) || (flag == null && user == null))
            {
               user = text;
               flag = null;

               continue;
            }

            // Authentication: password
            if ("-P".equals(flag) || (flag == null && password == null))
            {
               password = text;
               flag = null;

               continue;
            }

            // Socket service/port number.
            if ("-S".equals(flag) || (flag == null && port == null))
            {
               try
               {
                  // Argument could be a string representation of a port
                  // number, so try to parse it as such.
                  service = text;
                  port = new Integer(text);
               }
               catch (NumberFormatException exc)
               {
                  // Look up the port number associated with the given service
                  // name.
                  port = Utils.getPortForService(null, text);
                  if (port < 0)
                  {
                     throw new PersistenceException("Invalid service name:  " + text);
                  }
               }
               flag = null;
               
               continue;
            }
         }
         else if ("-S".equals(flag) || (flag == null && port == null))
         {
            port = num;
            flag = null;
            
            continue;
         }
      }
      
      if (host == null && port != null)
      {
         // TODO:  query host for current P2J server instance.
         throw new PersistenceException(
            "Database connect host must be specified for port '" + port + "'");
      }

      if (LOG.isLoggable(Level.INFO))
      {
         String msg = String.format(
            "{%s} Connecting database [PDB = %s; LDB = %s; HOST = %s; SVC = %s; " +
                                     "PORT = %d; UID = %s; PWD = %s]",
            Utils.describeContext(),
            pdb,
            ldb,
            host,
            service,
            port, 
            user, 
            (password == null) ? "null" : "***");
         LOG.log(Level.INFO, msg);
      }
      
      Database database = null;
      
      if (host != null)
      {
         database = new Database(pdb, host, port);
         InetSocketAddress address = database.getSocketAddress();
         if (SessionManager.get().isLocalNode(address))
         {
            database = new Database(pdb);
         }
      }
      else
      {
         database = new Database(pdb);
      }
      
      // TODO:  passing null for certAlias disables requester-side server
      // validation, which is fine for the common case.  To enable this, we
      // would need a directory lookup to find the correct certificate alias
      // for the remote server in the directory, based upon the IP address
      // and port (or service name) of the remote server.
      ldb = makeConnection(database, ldb, null); // if ldb == null, a valid ldb will be returned 
      if (ldb == null)
      {
         return; // connection failed
      }
      
      if (user == null && password == null)
      {
         // no user/password combination supplied, store the detected blank os username for userid
         setConnectUserImpl(pdb);
         return;
      }
      
      if (user == null || password == null)
      {
         // TODO: both error messages are printed on P4GL
         ErrorManager.throwError(1403,
                                 (user == null)
                                    ? "You have not supplied a parameter for argument -U"
                                    : "You have not supplied a parameter for argument -P",
                                 null);
         ErrorManager.throwError(325,
                                 "Unable to execute CONNECT statement",
                                 null);
         return;
      }
      
      // after a successful connection check if user/password match
      if (SecurityOps.getAuthLevel(pdb) != SecurityOps.FREE_ACCESS)
      {
         if (SecurityOps.setUserId(user, password, pdb).booleanValue())
         {
            // pin the authenticated used so it will be returned by DBPARAM function
            setConnectUserImpl(pdb);
         }
         else
         {
            // authentication failed
            ErrorManager.throwError(710,
                                    "Your Password and Userid " + user + " do not match",
                                    null);
         }
      }
      else 
      {
         ErrorManager.throwError(709,
                "-P and -U startup parameters require _User file records", 
                null);
      }
   }
   
   /**
    * Create a logical connection to the database with the specified physical
    * name, creating the specified logical name for it.
    *
    * @param   database
    *          Database information object.
    * @param   ldbName
    *          Logical database name.  If <code>null</code>, root name of
    *          physical database name will be used.
    * @param   certAlias
    *          Alias of certificate to be read from trust store to enable
    *          certificate-based, server-to-server authentication.  This
    *          should match the server ID of the remote node.
    *
    * @return  the logical database name on success (the ldbName input parameter may be 
    *          <code>null</code>, but a new logical name will be computed and returned) 
    *          and <code>null</code> otherwise.
    *
    * @throws  ErrorConditionException
    *          if the given database already is connected.
    * @throws  PersistenceException
    *          if there is an error registering the database with the
    *          database manager or establishing a network session with the
    *          remote server.
    */
   private String makeConnection(Database database, String ldbName, String certAlias)
   throws PersistenceException
   {
      // Normalize names.
      if (ldbName == null)
      {
         // if ldbName is null, root name of physical database name will be used.
         ldbName = ConnectionManager.normalizeFilename(database.getName());
      }
      ldbName = ldbName.toLowerCase();

      String pdb = normalizeFilename(database.getName());
      
      // Is this database awaiting disconnection?
      boolean disconn = pendingDisconn.contains(database) ||
                        pendingTxDisconn.contains(database);
      
      if (connected.containsKey(ldbName))
      {
         String existing = normalizeFilename(connected.get(ldbName).getName());
         
         // Warn user if database is already connected (and is not pending a disconnect).
         if (existing.equals(pdb))
         {
            if (!disconn)
            {
               ErrorManager.recordOrShowError(
                  1012,
                  "Warning -- database " + ldbName + " is already connected",
                  true);
               
               return null;
            }
         }
         // Error out if attempting to assign logical name to more than one physical database.
         else
         {
            int[] num = new int[] { 1012, 1018 };
            String[] msg = new String[]
            {
               "Warning -- database " + ldbName + " is already connected",
               "Database physical names " + pdb + "  and " + existing +
               " do not match.  No connect."
            };
            ErrorManager.recordOrShowError(num, msg, true);
            
            return null;
         }
      }
      
      // If this database is pending a disconnect, the connect and pending
      // disconnect cancel each other out.  If not, register it and remember it is connected.
      if (!pendingDisconn.remove(database) && !pendingTxDisconn.remove(database))
      {
         boolean local = database.isLocal();
         
         try
         {
            Session session = null;
            if (!local)
            {
               // Establish network session with remote server.
               SessionManager sessMgr = SessionManager.get();
               InetSocketAddress address = database.getSocketAddress();
               session = sessMgr.connectVirtual(address, certAlias, null);
            }

            sessions.put(database, session);
         }
         catch (Exception exc)
         {
            throw new PersistenceException(exc);
         }
         
         // Register database with database manager if not local or have not been registered already.
         // If the database is local but not "load_at_startup" it must be registered.
         // If it was already registered false will be returned
         local |= DatabaseManager.registerDatabase(database);
         
         if (local)
         {
            // Store mapping of logical name to database.
            connected.put(ldbName, database);
            userIds.put(ldbName, SecurityOps.getDefaultUserid(ldbName));

            if (LOG.isLoggable(Level.INFO))
            {
               String msg = String.format(
                  "{%s} Connected database '%s' [%s]",
                  Utils.describeContext(),
                  ldbName,
                  database.toString());
               LOG.log(Level.INFO, msg);

               if (LOG.isLoggable(Level.FINE))
               {
                  msg = String.format(
                     "{%s} Current connections:  [%s]",
                     Utils.describeContext(),
                     connected);
                  LOG.log(Level.FINE, msg);
               }
            }
         }
         else if (LOG.isLoggable(Level.WARNING))
         {
            String msg = String.format(
                  "{%s} Unable to make database connection for '%s' [%s]",
                  Utils.describeContext(),
                  ldbName,
                  database.toString());
            LOG.log(Level.WARNING, msg);
         }
      }
      else
      {
         // Store mapping of logical name to database.
         connected.put(ldbName, database);
         userIds.put(ldbName, SecurityOps.getDefaultUserid());

         if (LOG.isLoggable(Level.INFO))
         {
            String msg = String.format(
               "{%s} Cancelled disconnect of database '%s' [%s]",
               Utils.describeContext(),
               ldbName,
               database.toString());
            LOG.log(Level.INFO, msg);

            if (LOG.isLoggable(Level.FINE))
            {
               msg = String.format(
                  "{%s} Current connections:  [%s]",
                  Utils.describeContext(),
                  connected);
               LOG.log(Level.FINE, msg);
            }
         }
      }
      
      // return the used logical name for the connected database 
      return ldbName;
   }
   
   /**
    * Request a disconnect from a logical database.  This does not actually
    * perform the disconnect immediately, but registers <code>ldbName</code>
    * for disconnect at the appropriate time:  once no buffers reference the
    * database, and at the end of the current transaction (if any).
    *
    * @param   ldbName
    *          Logical name of the database to disconnect.
    *
    * @throws  PersistenceException
    *          if the specified database has not been opened.
    */
   private void disconnectImpl(String ldbName)
   throws PersistenceException
   {
      String ldb = getLDBName(ldbName);
      if (ldb == null)
      {
         throw new PersistenceException(
            "DISCONNECT:  Database " + ldbName + " is not open or unknown",
            1009);
      }
      
      // Remove the database from the connected map whether or not the
      // disconnect can occur immediately.
      Database database = connected.remove(ldb);
      userIds.remove(ldb); // keep in sync
      connectTimeUserid.remove(ldb);  // keep in sync
      
      // If any record buffers still reference the database, defer the
      // disconnect until all affected buffer scopes have closed.
      if (pdbReferences.containsKey(database))
      {
         pendingDisconn.add(database);
      }
      else
      {
         disconnectImmediately(database);
      }
   }
   
   /**
    * Tests whether a logical connection to a database exists in this context,
    * given a logical database name or alias.
    *
    * @param   name
    *          Logical database name or alias.
    *
    * @return  <code>true</code> if a logical connection exists;  else
    *          <code>false</code>.
    */
   private logical connectedImpl(String name)
   {
      // Get the logical DB name.
      String ldb = getLDBName(name);
      
      Database database = connected.get(ldb);
      boolean isConn = (database != null                   &&
                        !pendingDisconn.contains(database) &&
                        !pendingTxDisconn.contains(database));

      if (LOG.isLoggable(Level.INFO))
      {
         String ctx = Utils.describeContext();
         
         String msg = String.format(
            "{%s} Database '%s' (%s [%s]) connected status:  %s",
            ctx,
            name,
            ldb,
            (database != null ? database.toString() : null),
            String.valueOf(isConn));
         LOG.log(Level.INFO, msg);

         if (LOG.isLoggable(Level.FINE))
         {
            msg = String.format("{%s} connected:  %s", ctx, connected.toString());
            LOG.log(Level.FINE, msg);
            msg = String.format("{%s} pendingDisconn:  %s", ctx, pendingDisconn.toString());
            LOG.log(Level.FINE, msg);
            msg = String.format("{%s} pendingTxDisconn:  %s", ctx, pendingTxDisconn.toString());
            LOG.log(Level.FINE, msg);
         }
      }
      
      return (new logical(isConn));
   }
   
   /**
    * Create an alias for a logical database name. The alias may not conflict
    * with an existing logical database name (except for that of the database
    * which it aliases).  The alias will be created regardless of whether the
    * target database currently is connected;  however, an error will be
    * raised (or silently registered) in the event it is not, after the
    * alias has been created.  The alias persists until explicitly deleted
    * with one of the variants of the <code>deleteAlias</code> method, or
    * until the current session ends.
    *
    * @param   alias
    *          Alias to be created.
    * @param   ldbName
    *          Logical name of the database to which the alias is assigned.
    *
    * @throws  PersistenceException
    *          if the alias conflicts with an existing logical database name,
    *          other than <code>ldbName</code>.
    * @throws  ErrorConditionException
    *          if the target database is not connected, and we are not in
    *          silent error mode.
    */
   private void createAliasImpl(String alias, String ldbName)
   throws PersistenceException
   {
      // Normalize names.
      alias = alias.toLowerCase();
      ldbName = ldbName.toLowerCase();
      
      // Alias cannot be an existing logical database name (except for the
      // one it aliases).
      if (connected.containsKey(alias) && !alias.equals(ldbName))
      {
         throw new PersistenceException(
            "Alias " + alias + " conflicts with a database name",
            389);
      }
      
      aliases.put(alias, ldbName);
      
      // Logical database name should already exist.  Otherwise, it is an
      // error.  However, if in silent error mode, the alias will still be
      // created (note that this is different than silent error mode for
      // most other statements).
      if (!connected.containsKey(ldbName))
      {
         ErrorManager.recordOrThrowError(
            1660,
            "Logical dbname " + ldbName +
            " must be connected in order to add alias " + alias);
      }
   }
   
   /**
    * Delete the specified alias immediately.
    *
    * @param   alias
    *          Alias to be deleted.
    */
   private void deleteAliasImpl(String alias)
   {
      aliases.remove(alias.toLowerCase());
   }
   
   /**
    * Retrieve the alias at the specified index in the current context.
    *
    * @param   index
    *          One-based index of the alias (aliases are stored in the order
    *          in which they were created).
    *
    * @return  The target alias, or unknown value if <code>index</code> is
    *          invalid.
    */
   private character aliasImpl(int index)
   {
      String key = (String) iterateToIndex(aliases.keySet(), index);
      String alias = aliases.get(key);
      
      if (alias != null)
      {
         return (new character(alias));
      }
      
      // Return unknown value by default.
      return (new character());
   }

   /**
    * Implementation of the <code>DBPARAM()</code> Progress function.
    * Returns, as a character string, a comma-separated list of the parameters
    * used to connect to the database.
    *
    * <p>The method do a lookup on the database's name, or id it is null, the
    * alias will extracted using the index.
    * 
    * <p> Parameters returned by this method generally will not match those 
    * returned by the original 4GL implementation, since we don't support any 
    * others than those handled by <code>ConnectionManager.connect()</code>.
    *
    * @param   nameOrAlias
    *          Logical database name or alias.
    *
    * @param   index
    *          One-based index of the alias (aliases are stored in the order
    *          in which they were created).
    *
    * @return  A comma-separated list of the parameters used to connect to 
    *             the database
    */
   private character dbParamImpl(String nameOrAlias, int index)
   {
      if (nameOrAlias == null)
      {
         String key = (String) iterateToIndex(aliases.keySet(), index);
         String alias = aliases.get(key);

         if (alias == null)
         {
            return new character(alias);
         }
         nameOrAlias = alias;
      }

      // Get the logical DB name.
      String ldb = getLDBName(nameOrAlias);
      Database db = connected.get(ldb);

      if (db == null)
      {
         return new character();
      }

      StringBuilder sb = new StringBuilder();
      sb.append("-db ").append(db.getName());
      sb.append(",-dt ").append(ConnectionManager.dbType(nameOrAlias));
      sb.append(",-ld ").append(nameOrAlias);
      if (!db.isLocal())
      {
         sb.append(",-H ").append(db.getSocketAddress().getHostName());
         sb.append(",-S ").append(db.getSocketAddress().getPort());
      }
      
      String userID = connectTimeUserid.get(ldb);
      if (userID != null && !userID.isEmpty())
      {
         sb.append(",-U ").append(userID).append(",-P");
      }

      // -B 200
      // -L 10000
      // any other connection params not supported (yet)

      return new character(sb.toString());
   }
}
